#!/usr/bin/perl
#
# rndCiv.pl - make randomized FreeCiv rulesets
# $Id: rndCiv.pl,v 1.32 2003/10/11 14:24:33 banjo Exp $
# Copyright (C) 2003, Billy Naylor <banjo@actrix.co.nz>
# 
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.  See <http://www.fsf.org/copyleft/gpl.txt>.
# 
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
#

# -----------------------------------------------------------
# Usage

my $helptext = <<EOF;

$0: make randomized FreeCiv rulesets
Usage: $0 [OPTIONS] packagelist

Options:

    -v/-vv/-vvv  verboseness, more 'v's more verbose
    -s/--size    number of techs to have
    -o/--outdir  directory to write to
                 defaults to ./RndCiv/
    -d/--data    data directory
    -t/--techs   must have techs
                 ( use '-' to continue to package list )
    -h/--help    print this help & exit
    -l/--list    list known packages

EOF

# -----------------------------------------------------------
# Configuration

use strict;
srand (time ^ $$ ^ unpack "%L*", `ps axww | gzip`);

our $debug = 0;

# maxiumum number of techs freeciv allows
my $MAXTECHS = 199;

# How many techs
my $maxtechs = 90;

# data directory
my $datadir = './rulesets';

# output directory
my $outdir = './output';

# -----------------------------------------------------------
# Parse args

my ( @pkgs, $techs, @musthave );

while ( @ARGV ) {
	$_ = shift;

	if ( m/^-+h(elp)?$/ ) {
		print $helptext;
		exit 1;
	}

	elsif ( m/^-+s(ize)?$/ ) {
		die "No size value\n" unless scalar @ARGV;
		my $size = shift;
		die "Unable to have size '$size'\n" unless $size =~ /^\d+$/;
		$maxtechs = $size;
	}

	elsif ( m/^-+(v+)$/ ) {
		$debug = length $1;
		$debug and print "Setting Debug level to $debug\n";
	}

	elsif ( m/^-+t(echs)?$/ ) {
		while ( @ARGV ) {
			$_ = shift;
			last if ( m/^-+$/ );
			push @musthave, $_;
		}
	} 

	elsif ( m/^-+o(outdir)?$/ ) {
		die "No output directory\n" unless scalar @ARGV;
		$outdir = shift;
	}

	elsif ( m/^-+d(ata)?$/ ) {
		die "No data directory given\n" unless scalar @ARGV;
		$_ = shift;
		die "Unable to locate data dir '$_'\n"
		 unless ( -d $_ );
		$datadir = $_;
	}

	elsif ( m/^-+l(ist)?$/ ) {
		die "No data directory found\n" unless ( -d $datadir );
		print join ', ', map{ m/\.(\w+)$/; $1 } qx{ ls $datadir/techs.* };
		print "\n";
		exit 1;
	}

	elsif ( m/^-/ ) {
		next if m/^-+-+$/;
		die "Unknown option '$_'\n";
	}

	else { # packages
		die "Bad package name '$_'\n"
		unless m/^\w+$/;
		push @pkgs, $_;
	}
}

# -----------------------------------------------------------
# Parse tech packages


=pod $techs

  $techs->{
	$techname {
		{flags},
		{prereqs},
	},
  }

=cut

my ( $pkgseen, $pastTop, $pkgcount ) = ( {}, 0, 0 );
push @pkgs, '-';  # top of stack
while ( scalar @pkgs ) {
	my $pkg = shift @pkgs;

	if ( $pkg =~ m/^\-$/ ) {
		$pastTop = 1;
		next;	
	}

	my $filename = "$datadir/techs.$pkg";
	die "Unknown package '$pkg'\n"
	  unless ( -e $filename );
	next if ( defined $pkgseen->{$pkg} );
	$pkgseen->{$pkg} = ++$pkgcount;

	open( TECHS, "< $filename" )
	  or die "Unable to open package $pkg '$filename':$!\n";

	my $linecount = 0;
	for (<TECHS>) {
		$linecount++;
		next if m/^#/;      # ignore comments ( hash in column 1 )
		next if m/^\s*$/;   # and blank lines
		my $errhelp = "$_ [$filename : $linecount]";

		# grab commands
		if ( m/^\+\s+(\w+)\s+(.*?)\s*$/ ) {
			my( $cmd, $arg ) = ( $1, $2 );

			if ( $cmd eq 'read' and $arg =~ m/^\w+$/ ) {
				push @pkgs, $arg;
			}

			elsif ( $cmd eq 'choose' ) {
				next if $pastTop;
				my @choices = &shuffle( split /,\s*/, $arg );
				( scalar @choices ) or
				 warn "No options listed $errhelp\n" and next;
				push @musthave, shift @choices;
			}

			else {
				warn "Bad command: $errhelp\n";
			}
			next;
		}

		# grab techs
		next unless m/
			^(.*?)			# name
			\:\s+			# name delim
			(\((.*?)\))?	# possible flags
			\s*(.*?)\s*$	# requirements
		/x;
		my ( $name, $flags, @reqs ) = ( $1, $3, split( /\s*,\s*/, $4 ) );
		( scalar @reqs > 1 )
		 or die "Too few prereqs for tech '$name' $errhelp\n";
	
		if ( defined $techs->{$name} ) {
			$techs->{$name}{flags} = $flags
			 if rand() > 0.4;
		}
		else {
			$techs->{$name}{flags} = $flags;
		}
		push @{ $techs->{$name}{prereqs} }, @reqs;
	}
	close TECHS;
}

my @tek = keys %{ $techs };
die "No tech packages loaded\n"
 unless ( scalar @tek );

$maxtechs = ( scalar @tek ) if ( scalar @tek < $maxtechs );
$maxtechs = $MAXTECHS if ( $maxtechs > $MAXTECHS );

# -----------------------------------------------------------
# Build techtree

=pod $techtree

 $techtree->{
	'None' => 1,
	'Name' => $name,
 }

=cut

my $techtree  = { None => { req1 => undef, req2 => undef, node => rndCiv::node->new( 'None' ) } };
my $techcount = 0;

for my $tech ( @musthave ) {
	$debug == 1 and print "Must have tech $tech\n";
	&buildown( $tech )
	 or $debug and print "Unable to build tree down for tech '$tech'\n";
}

for my $tech ( &shuffle( @tek ) ) {
	last if ( $techcount > $maxtechs );
	$debug == 1 and print "Building down from $tech\n";
	&buildown( $tech )
	 or $debug and print "Unable to build tree down for tech '$tech'\n";
}

# -----------------------------------------------------------
# Write our ruleset

unless ( -d $outdir ) {
	mkdir $outdir
	 or die "Unable to make output directory '$outdir':$!\n";
}

open ( FILE, "> $outdir/techs.ruleset" )
 or die "Unable to open '$outdir/techs.ruleset' for writing:$!\n";

print FILE <<EOF;
; created by rndCiv.pl
; total $techcount techs

[datafile]
description="Randomized technology data for Freeciv"
options="1.9"

EOF

my $summary = {
	'None' => {
		pre       => [],
		units     => [],
		buildings => [],
		govs      => [],
	},
};
for my $name ( sort keys %{ $techtree } ) {
	next if $name eq 'None' or $name eq '';
	my ( $req1, $req2, $flags ) = (
		$techtree->{$name}{req1},
		$techtree->{$name}{req2},
		$techs->{$name}{flags}
	);

	my $advance = lc "advance_$name";
	$advance =~ s/ /_/g;

	( $req1, $req2 ) = ( $req2, $req1 )
	 if ( $req1 eq 'None' and $req2 ne 'None' );

	$flags ||= '""';

	print FILE "[$advance]\n"
	 . "name     = _(\"$name\")\n"
	 . "req1     = \"$req1\"\n"
	 . "req2     = \"$req2\"\n"
	 . "flags    = $flags\n\n";

	$summary->{$name}{pre}       = [ $req1, $req2 ];
	$summary->{$name}{units}     = [];
	$summary->{$name}{buildings} = [];
	$summary->{$name}{govs}      = [];
}

close FILE;

# -----------------------------------------------------------
# parse units rules

$debug > 1 and print "Parsing Units rules\n";
{
	my ( $units, $options, @unitnames ) = readData(
		'units',
		sort{ $pkgseen->{$b} <=> $pkgseen->{$a} } keys %{ $pkgseen },
	);

	# check obsolete by
	for my $u ( @unitnames ) {
		my ( $canbenone, $ob ) = ( 0, '' );
		for my $r ( shuffle( @{ $units->{$u}{Obsolete} } ) ) {
			$r eq 'None' and $canbenone = 1 and next;
			next unless exists $units->{$r};
			next if $units->{$r}{nogood} == 1;
			$ob = $r;
			last;
		}

		# early unit could be obsoleted by later unit that doesn't make it??
		if ( $ob eq '' ) {
			unless ( $canbenone ) {
				$units->{$u}{nogood} = 1;
				next;
			}
			$ob = 'None';
		}
		$units->{$u}{data} =~ s/^obsolete_by.*$/obsolete_by    = "$ob"/m;
		$debug and print "Unit $u obsolete by $ob\n";
	}

	my ( $partisan_req, $partisan_count ) = ( '', 0 );
	for my $u ( shuffle( @{ $options->{partisan_req} } ) ) {
		next unless exists $techtree->{$u};
		$partisan_req .= "\"$u\",";
		$partisan_count++;
		last if $partisan_count == 2;
	}
	$partisan_req =~ s/,$//;

	# print to file
	open ( FILE, "> $outdir/units.ruleset" )
	 or die "Unable to open '$outdir/units.ruleset' for writing:$!\n";

	print FILE <<EOF;
; created by rndCiv.pl

[datafile]
description="Randomized units data for Freeciv"
options="1.9"

[units_adjust]
max_hitpoints=0                 ; 0 means no max
max_firepower=0                 ; 0 means no max
firepower_factor=1              ; convenient for Civ1

EOF

	for my $u ( sort { $units->{$a}{level} <=> $units->{$b}{level}  } @unitnames ) {
		next if $units->{$u}{nogood} == 1;
		print FILE $units->{$u}{data} . "\n";
		push @{ $summary->{ $units->{$u}{pre} }{units} }, $u;
	}

	print FILE <<EOF;
[u_specials]
partisan_req = $partisan_req
EOF

	close FILE;
}
 
# -----------------------------------------------------------
# parse buildings rules

$debug and print "Parsing Buildings rules\n";
{
	my ( $buildings, $options, @buildingnames ) = readData(
		'buildings',
		sort{ $pkgseen->{$b} <=> $pkgseen->{$a} } keys %{ $pkgseen },
	);

	# check required building
	for my $u ( @buildingnames ) {
		my ( $canbenone, $req ) = ( 0, '' );
		for my $r ( shuffle( @{ $buildings->{$u}{BuildReq} } ) ) {
			$r eq 'None' and $canbenone = 1 and next;
			next unless exists $buildings->{$r};
			next if $buildings->{$r}{nogood} == 1;
			$req = $r;
			last;
		}

		# early building could be obsoleted by later building that doesn't make it??
		if ( $req eq '' ) {
			unless ( $canbenone ) {
				$buildings->{$u}{nogood} = 1;
				next;
			}
			$req = 'None';
		}
		$buildings->{$u}{data} =~ s/^bldg_req.*$/bldg_req      = "$req"/m;
	}

	# check obsolete_by
	for my $u ( @buildingnames ) {
		next if $buildings->{$u}{nogood} == 1;
		my ( $canbenone, $ob ) = ( 0, '' );
		for my $r ( shuffle( @{ $buildings->{$u}{Obsolete} } ) ) {
			$r eq 'None' and $canbenone = 1 and next;
			next unless exists $buildings->{$r};
			next if $buildings->{$r}{nogood} == 1;
			$ob = $r;
			last;
		}

		if ( $ob eq '' ) {
			unless ( $canbenone ) {
				$buildings->{$u}{nogood} = 1;
				next;
			}
			$ob = 'None';
		}
		$buildings->{$u}{data} =~ s/^obsolete_by.*$/obsolete_by      = "$ob"/m;
	}


	# print to file
	open ( FILE, "> $outdir/buildings.ruleset" )
	 or die "Unable to open '$outdir/buildings.ruleset' for writing:$!\n";

	print FILE <<EOF;
; created by rndCiv.pl

[datafile]
description="Randomized buildings data for Freeciv"
options="1.10.1"

EOF

	for my $u ( @buildingnames ) {
		next if $buildings->{$u}{nogood};
		print FILE $buildings->{$u}{data} . "\n";
		push @{ $summary->{ $buildings->{$u}{pre} }{buildings} }, $u;
	}

	print FILE <<EOF;
; FIXME: remove all of the following when gen-impr implemented...

[b_special]

; Special values:

aqueduct_size=8;
sewer_size=12;

; Techs which modify building effects:

cathedral_plus="Theology"
cathedral_minus="Communism"
colosseum_plus="Electricity"
temple_plus="Mysticism"
EOF

	close FILE;
}

# -----------------------------------------------------------
# parse government rules

$debug and print "Parsing Government rules\n";
{
	my ( $govs, $options, @govnames ) = readData(
		'governments',
		sort{ $pkgseen->{$b} <=> $pkgseen->{$a} } keys %{ $pkgseen },
	);
	
	my $aigoal = '';
	for my $u ( reverse sort( shuffle( @{ $options->{ai_goal} } ) ) ) {
		$u =~ s/^\s*\d+\s+//;
		next unless exists $govs->{$u} and not $govs->{$u}{nogood};
		$aigoal = $u;
		last;
	}
	die "Unable to assign ai_goal\n" unless $aigoal;

	my $ai_tech_hints = '';
	for my $u ( @{ $options->{ai_tech_hints} } ) {
		$u =~ m/^\s*(.+?),\s*(\d+),\s*(\d+),\s*(\d+),\s*(\d+)\s*$/
		 or warn "Invalid ai_tech_hints '$u'\n" and next;
		next unless exists $techtree->{$1};
		$ai_tech_hints .= "    \"$1\", $2, $3, $4, $5\n";	
	}

	# print to file
	open ( FILE, "> $outdir/governments.ruleset" )
	 or die "Unable to open '$outdir/governments.ruleset' for writing:$!\n";

	print FILE <<EOF;
; created by rndCiv.pl

[datafile]
description="Randomized governments data for Freeciv"
options="1.9"

[governments]
default="Despotism"
when_anarchy="Anarchy"
ai_goal="$aigoal"

ai_tech_hints = { "tech", "turns_factor", "const_factor", "get_first", "done"
$ai_tech_hints}
EOF

	for my $u ( @govnames ) {
		next if $govs->{$u}{nogood};
		print FILE $govs->{$u}{data} . "\n";
		push @{ $summary->{ $govs->{$u}{pre} }{govs} }, $u;
	}
	close FILE;
}

# -----------------------------------------------------------
# write summary

open ( FILE, "> $outdir/ruleset.summary" )
 or die "Unable to open '$outdir/ruleset.summary' for writing:$!\n";

print FILE "Summary of rndCiv ruleset"
 . "\nDate: " . scalar localtime( time )
 . "\n$techcount techs\n";

for my $t ( sort keys %{ $summary } ) {
	print FILE "\n$t (" . join ', ', @{ $summary->{$t}{pre} };
	print FILE ")";

	scalar @{ $summary->{$t}{units} } and
	print FILE "\n units: ", join ', ', @{ $summary->{$t}{units} };
	scalar @{ $summary->{$t}{buildings} } and
	print FILE "\n buildings: ", join ', ', @{ $summary->{$t}{buildings} };
	scalar @{ $summary->{$t}{govs} } and
	print FILE "\n governments: ", join ', ', @{ $summary->{$t}{govs} };
}

print FILE "\n\n";
close FILE;

exit 1;

# -----------------------------------------------------------
# subs

# The Fisher Yates Shuffle from the Perl Cookbook
sub shuffle {
	my @array = @_;
	for my $i ( 0..$#array ) {
		my $j = int rand( $i + 1 );
		next if $i == $j;
		( $array[$i], $array[$j] ) = ( $array[$j], $array[$i] );
	}
	return @array;
}

# begin to build down to 'None' from a given tech
sub buildown {
	my $tech = shift;
	my ( $tree, %tekList ) = &buildnodesdown( $tech, () )
	 or return 0;
	&readtree( $tree )
	 or return 0;
	return 1;
}

# create a tech tree down to 'None'
# might redefine working techs ??
sub buildnodesdown {
	my ( $name, %tekList ) = @_;

	if ( exists $techtree->{$name} ) {
		$debug > 2 and print "'$name' is in techtree\n";
		return $techtree->{$name}{node};
	}

	unless ( exists $techs->{$name} ) {
		$debug > 1 and print "Unknown tech '$name' called\n";
		return undef;
	}

	# already found 
	if ( exists $tekList{$name} ) {
		$debug > 2 and print "already have $name\n";
		return $tekList{$name}{node};
	}
	$tekList{$name}{node} = undef;

	# choose an available prereq
	my ( $pre1, $pre1name, $hold, $holdname );
	for my $p ( &shuffle( @{ $techs->{$name}{prereqs} } ) ) {
		$debug > 1 and print "$name: Trying prereq $p\n";
		my ( $ret, %newTL ) = &buildnodesdown( $p, %tekList );

		unless ( defined $ret ) {
			$debug > 2 and print "No path found for '$p'\n";
			next;
		}
		next if $ret->contains( $name );

		unless ( $pre1 ) {
			$pre1 = $ret;
			$pre1name = ( ref $pre1 ) ? $pre1->{name} : $pre1;
			next;
		}
		my $pre2 = $ret;
		my $pre2name = ( ref $pre2 ) ? $pre2->{name} : $pre2;

		unless ( $pre1name eq 'None' or $pre2name eq 'None' ) {
			next if ( $pre2name eq $pre1name );
			# i would like to use these constraints
			# but then the tree may not even build!
#			next if ( $pre1->{req1}{name} eq $pre2name );
#			next if ( $pre1->{req2}{name} eq $pre2name );
#			next if ( $pre2->{req1}{name} eq $pre1name );
#			next if ( $pre2->{req2}{name} eq $pre1name );
		}

		if ( $pre2name eq "None" ) {
			#next if ( $holdname ne 'None' );
			( $hold,  $holdname ) = ( $pre2, $pre2name );
			next;
		}
		
		( $hold,  $holdname ) = ( $pre2, $pre2name );
		if ( $pre1->contains( $pre1name )
		  or $pre2->contains( $pre2name ) ) {
			$debug > 2 and print "Holding $pre2name\n";
			next;
		}

		last;
	}

	unless ( $pre1 and $hold ) {
		$debug > 1 and print "Failed to find 2 prereqs for tech '$name'\n";
		return undef
	}

	$debug > 2 and print "New node for $name( $pre1name, $holdname )\n";
	my $node = rndCiv::node->new( $name );
	$node->{req1} = $pre1;
	$node->{req2} = $hold;
	return ( $node, %tekList );
}

# parse sucessful tree into $techtree
sub readtree {
	my $node = shift;
	my $name = ( ref $node ) ? $node->{name} : $node;
	return $name
	  if ( exists $techtree->{$name} );

	$techtree->{$name} = {};
	$techtree->{$name}{node} = $node;
	for my $r ( 'req1', 'req2' ) {
		$techtree->{$name}{$r} = ( ref $node->{$r} ) ?
		 &readtree( $node->{$r} ) : $node->{$r};
	}
	$techcount++;
	return $name;
}

# utility for reading governments, buildings and units data files
sub readData {
	my ( $type, @pkgList ) = @_;
	die "Bad type '$type'\n"
	 unless ( $type =~ m/^(units|governments|buildings)$/ );

	my ( @list, $options ) = ( (), {} );
	for my $p ( @pkgList ) {
		$debug > 1 and print "Reading package '$datadir/$type.$p'\n";
		unless( open( FILE, "< $datadir/$type.$p" ) ) {
			$debug > 2 and print "units ruleset $datadir/$type.$p not found\n";
			next;
		}

		my ( $buf, $level ) = ( '', 0 );
		for (<FILE>) {
			if ( m/^\[/ and $buf ) {
				push @list, rndCiv::Data->new( $buf, $level ); # level is for ordering units
				$buf = '';
			}
			elsif ( m/^\+\s+level\s+(\d+)/ ) {
				$level = $1;
				next;
			}
			elsif ( m/^\+\s+(\w+)\s+(.*?)\s*$/ ) {
				push @{ $options->{$1} }, $2;
				next;
			}
			next if m/^\#/;
			next if m/^\s*$/;
			next if m/^\s*\;/;
			$buf .= $_;
		}
		close FILE;
		$buf and push @list, rndCiv::Data->new( $buf, $level );
	}

	my ( $data, @names );
	for my $ob ( @list ) {
		$debug and print "Checking $type $ob->{Name}\n";
		my $pre = '';
		for my $r ( shuffle( @{ $ob->{PreReqs} } ) ) {
			next unless exists $techtree->{$r};
			$pre = $r;
		}
		unless ( $pre ) {
			$debug and print "No Prereqs found for $type '$ob->{Name}'\n";
			next;
		}
		$ob->{data} =~ s/^tech_req.*$/tech_req      = "$pre"/m;
		$ob->{pre} = $pre;
		
		# overwriting multiple defs
		$data->{ $ob->{Name} } = $ob;
		push @names, $ob->{Name};
	}

	return ( $data,  $options, @names );
}

# -----------------------------------------------------------
# object class for nodes

package rndCiv::node;

sub new {
	my ( $class, $name ) = @_;
	return bless {
		name => $name,
		req1 => undef,
		req2 => undef,
	}, $class;
}

sub contains {
	my ( $self, $name ) = @_;
	return 0 if $self->{name} eq 'None';
	return 1 if $self->{req1}->contains( $name );	
	return 1 if $self->{req2}->contains( $name );	
	return 0;
}

package rndCiv::Data;

sub new {
	my ( $class, $buf, $level ) = @_;

	my $self = bless {
		data     => $buf,
		Name     => '',
		PreReqs  => [],
		Obsolete => [],
		BuildReq => [],
		nogood   => 0,
		level    => $level,
	}, $class;

	$buf =~ m/^name\s+=\s+_\(\"(.+?)\"\)\s*$/m
	 or die "No name found for data '$buf'\n";
	$self->{Name} = $1;

	$buf =~ m/^tech_req\s+=\s+(.*?)\s*$/m
	  or die "No prereqs found for data '$self->{Name}'\n$buf\n";
	my $pre = $1;
	$pre =~ s/\"//g;
	push @{ $self->{PreReqs} }, split /\s*\|\s*/, $pre;

	if ( $buf =~ m/^obsolete_by\s+=\s+(.*?)\s*$/m ) {
		my $ob = $1;
		$ob =~ s/\"//g;
		push @{ $self->{Obsolete} }, split /\s*\|\s*/, $ob;
	}

	if ( $buf =~ m/^bldg_req\s+=\s+(.*?)\s*$/m ) {
		my $req = $1;
		$req =~ s/\"//g;
		push @{ $self->{BuildReq} }, split /\s*\|\s*/, $req;
	}

	return $self;
}

1;

